package org.bdware.sc;

import org.bdware.sc.bean.Contract;
import org.bdware.sc.bean.FunctionDesp;
import org.bdware.sc.bean.IDSerializable;
import org.bdware.sc.event.REvent;
import org.bdware.sc.node.AnnotationNode;
import org.bdware.sc.node.YjsType;
import org.bdware.server.trustedmodel.ContractExecutor;
import org.bdware.server.trustedmodel.SingleNodeExecutor;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ContractMeta implements IDSerializable {
    public Contract contract;
    public transient ContractExecutor contractExecutor;
    ContractStatusEnum status;
    String name;
    String id;
    boolean isDebug;
    Map<String, REvent.REventSemantics> declaredEvents;
    List<FunctionDesp> exportedFunctions;
    Map<String, String> logDetail;
    List<AnnotationNode> annotations;
    Set<String> dependentContracts;
    transient Map<String, FunctionDesp> funCache;
    boolean sigRequired;
    String thisPermission; // 合约当前权限
    /*
    {
       "name": "dx_substr",
       "parameter":
       {
           "columnIndex":5,
           "paras":["1","3"]
       }
    },
    */
    // Map<Object,Object>MaskInfo;

    public ContractMeta() {
        contractExecutor = SingleNodeExecutor.instance;
    }

    public ContractStatusEnum getStatus() {
        return status;
    }

    public void setStatus(ContractStatusEnum status) {
        this.status = status;
    }

    public String getID() {
        return id;
    }

    public List<AnnotationNode> getAnnotations() {
        return annotations;
    }

    public List<FunctionDesp> getExportedFunctions() {
        return exportedFunctions;
    }

    public Set<String> getDependentContracts() {
        return dependentContracts;
    }

    public Map<String, REvent.REventSemantics> getEvents() {
        return declaredEvents;
    }

    public YjsType getYjsType() {
        return contract.getYjsType();
    }

    public String getPubkey() {
        return contract.getPublicKey();
    }

    public String getThisPermission() {
        return thisPermission;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    // public setMask(){}

    public boolean getIsDebug() {
        return isDebug;
    }

    public FunctionDesp getExportedFunction(String action) {
        FunctionDesp ret;
        if (null == funCache) {
            funCache = new HashMap<>();
        }
        if (!funCache.containsKey(action)) {
            ret = seekFunction(action);
            funCache.put(action, ret);
        } else {
            ret = funCache.get(action);
        }
        return ret;
    }

    private FunctionDesp seekFunction(String action) {
        for (FunctionDesp desp : exportedFunctions) {
            if (desp != null && desp.functionName.equals(action)) return desp;
        }
        return null;
    }

    public void setContract(Contract c) {
        contract = c;
        id = c.getID();
        status = ContractStatusEnum.HANGED;
        isDebug = false;
    }

    public void setContractExecutor(ContractExecutor executor) {
        this.contractExecutor = executor;
    }
}
