package org.bdware.sc.consistency.pbft;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractManager;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.Node;
import org.bdware.sc.consistency.CommitAlgorithm;
import org.bdware.sc.consistency.Committer;
import org.bdware.sc.consistency.Pair;
import org.bdware.sc.units.TrustfulExecutorConnection;
import org.bdware.sc.util.JsonUtil;
import org.zz.gmhelper.SM2KeyPair;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

public class PBFTAlgorithm implements CommitAlgorithm {
    private static final Logger LOGGER = LogManager.getLogger(PBFTAlgorithm.class);

    static byte[] GETPUBKEY = "GETPUBKEY".getBytes();
    Committer committer;
    Map<Node, PBFTMember> members;
    AtomicLong allocatedID = new AtomicLong(0);
    Map<Long, PCInfo> info;
    List<PBFTMessage> commitedMsg;
    AtomicLong commitedOrder = new AtomicLong(0);
    Map<Integer, Pair<Node, PBFTMessage>> original;
    boolean isMaster;
    private TrustfulExecutorConnection connection;
    private int sendID;

    public PBFTAlgorithm(boolean isMaster) {
        commitedMsg = new ArrayList<>();
        info = new HashMap<>();
        members = new HashMap<>();
        original = new HashMap<>();
        this.isMaster = isMaster;
    }

    public void setSendID(int id) {
        sendID = id;
    }

    public void addMember(Node node, PBFTMember member) {
        members.put(node, member);
    }

    public void setCommitter(Committer c) {
        committer = c;
    }

    public void setConnection(TrustfulExecutorConnection c) {
        connection = c;
    }

    private void onPBFTMessage(final Node sender, final PBFTMessage pbftMessage) {
        // pbftMessage.type != Request
        // verify signed message
        // System.out.println(getPort() + " recv:" + pbftMessage.type + " len:" +
        // pbftMessage.content.length);
        PCInfo temp;
        PBFTMessage prepareMsg;
        LOGGER.info("recv: " + pbftMessage.getDisplayStr());
        switch (pbftMessage.type) {
            case AddMember:
                //  PBFTMember member = PBFTMember.parse(pbftMessage.content);
                //  if (member != null)
                //    members.put(sender, member);
                // case DeleteMember:
                // JUST ignore
                break;
            case Request:
                // all nodes save full content, clientIP, clientPort !!!!
                int hash = java.util.Arrays.hashCode(pbftMessage.content);
                original.put(hash, new Pair<Node, PBFTMessage>(sender, pbftMessage));
                if (isMaster) {
                    prepareMsg = new PBFTMessage();
                    prepareMsg.order = allocatedID.incrementAndGet();
                    prepareMsg.type = PBFTType.PrePrepare;
                    prepareMsg.content = (java.util.Arrays.hashCode(pbftMessage.content) + "").getBytes();

                    temp = new PCInfo();
                    temp.request = pbftMessage;
                    if (info.get(prepareMsg.order) != null) {
                        System.out.println("==========================Error!!!");
                    }
                    info.put(prepareMsg.order, temp);
                    temp.isPrePrepareReceived = true;
                    broadcast(prepareMsg);
                } else {
                    // System.out.println(getPort() + " Request:" + pbftMessage.getDisplayStr());
                    matchPrePrepareFromOriginReqeust(hash, pbftMessage);
                }
                reply(sender, "success");
                break;
            case PrePrepare:
                // assert !isMaster
                // Get Pubkey of Master
                // verify the content matches full content !!!! and client info
                if (info.get(pbftMessage.order) == null) {
                    info.put(pbftMessage.order, new PCInfo());
                }
                matchOriginalReqeustFromPrePrepare(pbftMessage);
                updatePrepare(info.get(pbftMessage.order));
                // TODO 触发prepare
                break;
            case Prepare:
                temp = info.get(pbftMessage.order);
                LOGGER.info("receive Prepare from:" + pbftMessage.sendID + " -> " + pbftMessage.order);
                if (temp == null) {
                    PCInfo pcInfo = new PCInfo();
                    pcInfo.buff.add(pbftMessage);
                    info.put(pbftMessage.order, pcInfo);
                    requestPrePrepareFromMaster(pbftMessage.order);
                } else if (temp.updatePrepare(pbftMessage, this)) {
                    // check the sender (is master) and order is in range !!!!
                    PBFTMessage commitMsg = new PBFTMessage();
                    commitMsg.type = PBFTType.Commit;
                    commitMsg.order = pbftMessage.order;
                    commitMsg.content = new byte[1];
                    updateCommit(commitMsg);
                    broadcast(commitMsg);
                }
                break;
            case Commit:
                temp = info.get(pbftMessage.order);
                if (temp == null) {
                    PCInfo pcInfo = new PCInfo();
                    pcInfo.buff.add(pbftMessage);
                    info.put(pbftMessage.order, pcInfo);
                    requestPrePrepareFromMaster(pbftMessage.order);
                    break;
                }
                if (temp.updateCommit(pbftMessage, this)) {
                    // check all messages' order !!!!
                    // write order to content !!!!
                    PBFTMessage commitMsg = new PBFTMessage();
                    commitMsg.type = PBFTType.Reply;
                    // execute in order
                    onCommit(pbftMessage);// once complete automatically send to client !!!!
                    // broadCast(commitMsg);
                }
                break;
            case Reply:// client receive this and check f+1 same result
                if (isMaster) {
                    System.out.println(pbftMessage.order + " " + pbftMessage.sendID);
                }
                break;
            case ReSend:
                // TODO
                if (isMaster) {
                    prepareMsg = new PBFTMessage();
                    prepareMsg.order = pbftMessage.order;
                    prepareMsg.type = PBFTType.PrePrepare;
                    temp = info.get(prepareMsg.order);
                    prepareMsg.content = (java.util.Arrays.hashCode(temp.request.content) + "").getBytes();
                    broadcast(pbftMessage);
                }
            default:
        }

        // printDebugInfo();

    }

    private void updateCommit(PBFTMessage commitMsg) {
        PCInfo temp = info.get(commitMsg.order);
        if (temp.updateCommit(commitMsg, this)) {
            // check all messages' order !!!!
            // write order to content !!!!
            PBFTMessage replyMsg = new PBFTMessage();
            commitMsg.type = PBFTType.Reply;
            // execute in order
            onCommit(commitMsg);// once complete automatically send to client !!!!
            // broadCast(commitMsg);
        }

    }

    private void updatePrepare(PCInfo pcInfo) {
        for (PBFTMessage msg : pcInfo.buff) {
            if (msg.type == PBFTType.Prepare) {
                if (pcInfo.updatePrepare(msg, this)) {
                    PBFTMessage commitMsg = new PBFTMessage();
                    commitMsg.type = PBFTType.Commit;
                    commitMsg.order = msg.order;
                    commitMsg.content = new byte[1];
                    updateCommit(commitMsg);
                    broadcast(commitMsg);
                }
            }
        }
    }

    private void requestPrePrepareFromMaster(long order) {

//        PBFTMessage message = new PBFTMessage();
//        message.type = PBFTType.ReSend;
//        message.order = order;
//        message.content = new byte[1];
//        sendToMaster(message);
        LOGGER.info("request Resend");
    }

    private void matchPrePrepareFromOriginReqeust(int hash, PBFTMessage pbftMessage) {
        for (PCInfo pcInfo : info.values()) {
            for (PBFTMessage msg : pcInfo.buff) {
                if (msg.type == PBFTType.PrePrepare && Integer.parseInt(new String(msg.content)) == hash) {
                    handlePrePrepare(msg, original.get(hash).second);
                    return;
                }
            }
        }
    }

    private void reply(Node receiver, String data) {

    }

    private void retryLater(int delay, final Node sender, final PBFTMessage pbftMessage) {
        ContractManager.scheduledThreadPool.schedule(
                () -> onPBFTMessage(sender, pbftMessage),
                delay,
                TimeUnit.MILLISECONDS);
    }

    private void handlePrePrepare(PBFTMessage pbftMessage, PBFTMessage req) {
        PBFTMessage prepareMsg = new PBFTMessage();
        prepareMsg.type = PBFTType.Prepare;
        prepareMsg.order = pbftMessage.order;
        prepareMsg.content = pbftMessage.content;
        PCInfo temp;
        if (info.get(prepareMsg.order) == null) {
            temp = new PCInfo();
            info.put(prepareMsg.order, temp);
        } else
            temp = info.get(prepareMsg.order);
        temp.request = req;
        temp.isPrePrepareReceived = true;
        temp.updatePrepare(prepareMsg, this);
        // System.out.println(getPort() + ":sendPrepare " + pbftMessage.order);
        broadcast(prepareMsg);

    }

    private void matchOriginalReqeustFromPrePrepare(PBFTMessage pbftMessage) {
        int receivedHash = Integer.parseInt(new String(pbftMessage.content));
        if (original.get(receivedHash) != null) {
            handlePrePrepare(pbftMessage, original.get(receivedHash).second);
        } else {
            info.get(pbftMessage.order).buff.add(pbftMessage);
        }
    }

    private synchronized void onCommit(PBFTMessage pbftMessage) {
        PBFTMessage original = getOriginalMessage(pbftMessage.order);
        // execute(pbftMessage);
        if (pbftMessage.order == commitedOrder.get() + 1) {
            execute(pbftMessage);
        } else {
            commitedMsg.add(original);
        }
    }

    private synchronized void execute(PBFTMessage pbftMessage) {
        commitedOrder.incrementAndGet();
        LOGGER.info("execute:" + pbftMessage + " infoKeySet:" + JsonUtil.toJson(info.keySet()));


        PBFTMessage original = getOriginalMessage(pbftMessage.order);
        ContractRequest msg = ContractRequest.parse(original.content);
        committer.onCommit(msg);

        info.remove(pbftMessage.order);
        // ContractProcess.instance.onEvent(msg);
        commitedMsg.sort((o1, o2) -> (int) (o1.order - o2.order));
        Set<PBFTMessage> executedMsgs = new HashSet<>();
        for (PBFTMessage message : commitedMsg) {
            if (commitedOrder.get() == message.order) {
                commitedOrder.incrementAndGet();
                msg = ContractRequest.parse(getOriginalMessage(message.order).content);
                committer.onCommit(msg);
                executedMsgs.add(message);
                // info.remove(pbftMessage.order);
            } else
                break;
        }
        commitedMsg.removeAll(executedMsgs);

    }

    private PBFTMessage getOriginalMessage(long order) {
        return info.get(order).request;
    }

    private void broadcast(PBFTMessage pbftMessage) {
        pbftMessage.sendID = sendID;
        connection.broadcast(pbftMessage.getBytes());
    }

    private void sendToMaster(PBFTMessage pbftMessage) {
        for (Node node : members.keySet()) {
            PBFTMember member = members.get(node);
            if (member.isMaster) {
                pbftMessage.sendID = sendID;
                connection.sendMessage(node, pbftMessage.getBytes());
                return;
            }
        }
    }

    public int getMemberSize() {
        return members.size();
    }

    @Override
    public void onMessage(Node node, byte[] msg) {
        PBFTMessage pbftMessage = PBFTMessage.parse(msg);
        onPBFTMessage(node, pbftMessage);
    }

    public void setAtomSeq(int i) {
        allocatedID.set(i);
        commitedOrder.set(i);
    }

    static class NodeInfo {
        SM2KeyPair privKey;
        int hash;
        boolean isMaster;

        public NodeInfo(SM2KeyPair key) {
            privKey = key;
            hash = Arrays.hashCode(privKey.getPublicKey().getQ().getEncoded(false));
            isMaster = false;
        }

        public int getNodeID() {
            return hash;
        }

        public boolean isMaster() {
            return isMaster;
        }

    }

    static class PCInfo {
        public PBFTMessage request;
        Set<Integer> prepare;
        Set<Integer> commit;
        boolean isPrePrepareReceived;
        boolean isSendCommit;
        boolean isSendReply;
        List<PBFTMessage> buff;

        PCInfo() {
            prepare = new HashSet<>();
            commit = new HashSet<>();
            buff = new ArrayList<>();
            isSendCommit = false;
            isSendReply = false;
            isPrePrepareReceived = false;
        }

        public synchronized boolean updatePrepare(PBFTMessage message, PBFTAlgorithm center) {
            if (isSendCommit)
                return false;
            prepare.add(message.sendID);
            if (!isPrePrepareReceived) {
                return false;
            }
//			System.out.println("---: updatePrepare, size:" + prepare.size() + " -->"
//					+ (center.members.size() / 3 * 2 + 1) + " --senderID:" + message.sendID);

            if (prepare.size() > center.members.size() / 3 * 2) {
                isSendCommit = true;
                return true;
            }
            return false;
        }

        public synchronized boolean updateCommit(PBFTMessage message, PBFTAlgorithm center) {
            if (isSendReply)
                return false;
            commit.add(message.sendID);
            if (!isSendCommit)
                return false;
            if (commit.size() > center.members.size() / 3 * 2) {
                prepare = null;
                commit = null;
                isSendReply = true;
                return true;
            }
            return false;
        }

        public String getDisplayStr() {
            return "pSize:" + (prepare == null ? "null" : prepare.size()) + " cSize:"
                    + (commit == null ? "null" : commit.size()) + " isSendCommit:" + isSendCommit + " isSendReply:"
                    + isSendReply + " buffSize:" + buff.size();
        }
    }
}
