package org.bdware.sc.consistency.pbft;

import org.bdware.sc.conn.ByteUtil;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

public class PBFTMessage {
	PBFTType type;
	int sendID;
	long order;
	byte[] content;
	String stamp;// time-stamp or other string, given by client, to identify the content
	String signature;

	public int getBytesHash() {
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
		bo.write(type.toInt());
		ByteUtil.writeLong(bo, sendID);
		ByteUtil.writeLong(bo, order);
		try {
			bo.write(content);
		} catch (IOException e) {
			e.printStackTrace();
		}
		byte[] ret = bo.toByteArray();
		int hashCode = ret.hashCode();
		return hashCode;
	}

	public byte[] getBytes() {
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
		try {
			bo.write(type.toInt());
			ByteUtil.writeInt(bo, sendID);
			ByteUtil.writeLong(bo, order);
			if (signature != null) {
				byte[] signature = this.signature.getBytes();
				ByteUtil.writeInt(bo, signature.length);
				bo.write(signature);
			} else
				ByteUtil.writeInt(bo, 0);

			bo.write(content);
		} catch (IOException e) {
			e.printStackTrace();
		}
		return bo.toByteArray();
	}

	public static PBFTMessage parse(byte[] b) {
		PBFTMessage msg = new PBFTMessage();
		ByteArrayInputStream bi = new ByteArrayInputStream(b);
		msg.type = PBFTType.fromByte(bi.read());
		msg.sendID = ByteUtil.readInt(bi);
		msg.order = ByteUtil.readLong(bi);
		int sigLen = ByteUtil.readInt(bi);
		if (sigLen > 0)
			msg.signature = new String(ByteUtil.readBytes(bi, sigLen));
		else
			msg.signature = null;
		msg.content = ByteUtil.readBytes(bi, b.length - 1 - 4 - 8 - 4 - sigLen);
		return msg;
	}

	private static String IDA = "PBFTMsg";

//	public void sign(SM2KeyPair pair) {
//		sendID = Arrays.hashCode(pair.getPublicKey().getEncoded(false));
//		signature = sm02.sign(content, IDA, pair).toString();
//	}
//
//	public boolean verify(PBFTMember member) {
//		return sm02.verify(content, Signature.loadFromString(signature), IDA,
//				SM2KeyPair.publicKeyStr2ECPoint(member.pubKey));
//	}

	public void setType(PBFTType type) {
		this.type = type;
	}

	public void setContent(byte[] content) {
		this.content = content;
	}

	public String getDisplayStr() {
		StringBuilder sb = new StringBuilder();
		sb.append(sendID);
		sb.append("_");
		sb.append(order);
		sb.append("_TYPE_").append(type);
		sb.append(" --> sig:");
		sb.append(signature);
		sb.append(" content:");
		sb.append(new String(content));
		return sb.toString();
	}

	public void setOrder(long order) {
		this.order = order;
	}
}
